/*
   From "Principles of Parallel Programming", pages 190-192.

   This program should be run eight times,
   with all eight combinations of
     a) myCount with and without volatile
     b) SPACING_FACTOR equal to 1 or 64
     c) with and without the JIT compiler


     https://stackoverflow.com/questions/4667483/how-is-the-default-java-heap-size-determined
     https://stackoverflow.com/questions/28272923/default-xmxsize-in-java-8

     java -XX:+PrintFlagsFinal -version | findstr HeapSize
     java -XX:+PrintCommandLineFlags -verion

      -Xms<size>   set initial Java heap size
      -Xmx<size>   set maximum Java heap size
      -Xss<size>   set java thread stack size
      -Xprof       output cpu profiling data
*/

import java.util.Random;

public class Count3s implements Runnable
{
   private static final Random random = new Random();

   private static final int ARRAY_LENGTH = (int)Math.pow(2, 30); // 2^30 = 1GB
   private static int[] array = new int[ARRAY_LENGTH];
   private static final int RANGE = 10;

   private static final int THREADS = 4; // be sure THREADS divides ARRAY_LENGTH
   private static final int SPACING_FACTOR = 64; // try 1 or 64
   private static volatile int[] myCount = new int[THREADS * SPACING_FACTOR]; // with volatile
 //private static int[] myCount = new int[THREADS * SPACING_FACTOR]; // without volatile
   private static int count = 0;

   private int startIndex;
   private int elements;
   private int threadIndex;

   public Count3s(int startIndex, int elements, int threadIndex)
   {
      this.startIndex = startIndex;
      this.elements = elements;
      this.threadIndex = threadIndex;
      myCount[threadIndex * SPACING_FACTOR] = 0;
   }


   public void run()
   {
      // count the number of threes
      for(int i = 0; i < elements; i++)
      {
         if(array[startIndex + i] == 3)
         {
            myCount[threadIndex * SPACING_FACTOR]++;
         }
      }
   }//run()


   public static void main(String[] args)
   {
      // initialize the elements in the data array
      for(int i = 0; i < array.length; i++)
      {
         array[i] = random.nextInt(RANGE);
      }

      // create the threads
      Count3s[] counters = new Count3s[THREADS];
      // be sure THREADS divides ARRAY_LENGTH
      int lengthPerThread = ARRAY_LENGTH / THREADS;

      for(int i = 0; i < counters.length; i++)
      {
         counters[i] = new Count3s(i*lengthPerThread, lengthPerThread, i);
      }

      // run the threads
      System.out.println("Starting threads.");
      Thread[] t = new Thread[counters.length];
      long start = System.currentTimeMillis();
      for(int i = 0; i < counters.length; i++)
      {
         t[i] = new Thread(counters[i]);
         t[i].start();
      }

      for(int i = 0; i < counters.length; i++)
      {
         try
         {
            t[i].join();
         }
         catch(InterruptedException e){}

         count += myCount[i * SPACING_FACTOR];
         //System.out.println("i = " + i + ", and myCount = " + myCount[i * SPACING_FACTOR]);
      }
      long end = System.currentTimeMillis();
      System.out.printf("total runtime: %,dms\n", end - start);

      // print the number of threes
      System.out.println("Number of threads: " + THREADS);
      System.out.println("SPACING_FACTOR: " + SPACING_FACTOR);
      System.out.printf("Number of 3's: %,d\n", count);

      System.out.println("\nSerial.");
      start = System.currentTimeMillis();
      count = 0;
      for(int i = 0; i < array.length; i++)
      {
         if(array[i] == 3)
         {
            count++;
         }
      }
      end = System.currentTimeMillis();
      System.out.printf("total runtime: %,dms\n", end - start);
      System.out.printf("Number of 3's: %,d\n", count);
      //System.out.println( java.util.Arrays.toString(array) );
      //System.out.println( java.util.Arrays.toString(myCount) );
   }//main()
}
